#######################################################
## .0. Mineral Abbreviations                     !!! ##
#######################################################
MINERAL_ABBREV_MAP: dict[str, dict[str, str]] = {
    "alm": {
        "name": "almandine",
        "group": "Garnets & olivines",
        "formula": "Fe3Al2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "andr": {
        "name": "andradite",
        "group": "Garnets & olivines",
        "formula": "Ca3Fe2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "gr": {
        "name": "grossular",
        "group": "Garnets & olivines",
        "formula": "Ca3Al2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "knor": {
        "name": "knorringite",
        "group": "Garnets & olivines",
        "formula": "Mg3Cr2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "maj": {
        "name": "majorite",
        "group": "Garnets & olivines",
        "formula": "Mg4Si4O12",
        "crystal_system": "",
        "other_info": "",
    },
    "py": {
        "name": "pyrope",
        "group": "Garnets & olivines",
        "formula": "Mg3Al2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "pyilm": {
        "name": "pyrope [ilmenite]",
        "group": "Garnets & olivines",
        "formula": "Mg3Al2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "spss": {
        "name": "spessartine",
        "group": "Garnets & olivines",
        "formula": "Mn3Al2Si3O12",
        "crystal_system": "",
        "other_info": "",
    },
    "chum": {
        "name": "clinohumite",
        "group": "Garnets & olivines",
        "formula": "Mg9Si4O16(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fa": {
        "name": "fayalite",
        "group": "Garnets & olivines",
        "formula": "Fe2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "fasp": {
        "name": "fayalite [spinel]",
        "group": "High-P phases",
        "formula": "Fe2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "fo": {
        "name": "forsterite",
        "group": "Garnets & olivines",
        "formula": "Mg2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "fosp": {
        "name": "forsterite [spinel]",
        "group": "High-P phases",
        "formula": "Mg2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "lrn": {
        "name": "larnite",
        "group": "Garnets & olivines",
        "formula": "Ca2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "mont": {
        "name": "monticellite",
        "group": "Garnets & olivines",
        "formula": "CaMgSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "teph": {
        "name": "tephroite",
        "group": "Garnets & olivines",
        "formula": "Mn2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "and": {
        "name": "andalusite",
        "group": "Aluminosilicates",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "ky": {
        "name": "kyanite",
        "group": "Aluminosilicates",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "sill": {
        "name": "sillimanite",
        "group": "Aluminosilicates",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "amul": {
        "name": "Al-mullite",
        "group": "Aluminosilicates",
        "formula": "Al2.5Si0.5O4.75",
        "crystal_system": "",
        "other_info": "",
    },
    "smul": {
        "name": "Si-mullite",
        "group": "Aluminosilicates",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "fctd": {
        "name": "Fe-chloritoid",
        "group": "Aluminosilicates",
        "formula": "FeAl2SiO5(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mctd": {
        "name": "Mg-chloritoid",
        "group": "Aluminosilicates",
        "formula": "MgAl2SiO5(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mnctd": {
        "name": "Mn-chloritoid",
        "group": "Aluminosilicates",
        "formula": "MnAl2SiO5(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fst": {
        "name": "Fe-staurolite",
        "group": "Aluminosilicates",
        "formula": "Fe4Al18Si7.5O44(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "mst": {
        "name": "Mg-staurolite",
        "group": "Aluminosilicates",
        "formula": "Mg4Al18Si7.5O44(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "mnst": {
        "name": "Mn-staurolite",
        "group": "Aluminosilicates",
        "formula": "Mn4Al18Si7.5O44(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "tpz": {
        "name": "hydroxy-topaz",
        "group": "Aluminosilicates",
        "formula": "Al2SiO4(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "ak": {
        "name": "akermanite",
        "group": "Other orthosilicates",
        "formula": "Ca2MgSi2O7",
        "crystal_system": "",
        "other_info": "",
    },
    "geh": {
        "name": "gehlenite",
        "group": "Other orthosilicates",
        "formula": "Ca2Al2SiO7",
        "crystal_system": "",
        "other_info": "",
    },
    "jgd": {
        "name": "julgoldite [FeFe]",
        "group": "Other orthosilicates",
        "formula": "Ca4Fe6Si6O21(OH)7",
        "crystal_system": "",
        "other_info": "",
    },
    "merw": {
        "name": "merwinite",
        "group": "Other orthosilicates",
        "formula": "Ca3MgSi2O8",
        "crystal_system": "",
        "other_info": "",
    },
    "mpm": {
        "name": "pumpellyite [MgAl]",
        "group": "Other orthosilicates",
        "formula": "Ca4MgAl5Si6O21(OH)7",
        "crystal_system": "",
        "other_info": "",
    },
    "fpm": {
        "name": "pumpellyite [FeAl]",
        "group": "Other orthosilicates",
        "formula": "Ca4FeAl5Si6O21(OH)7",
        "crystal_system": "",
        "other_info": "",
    },
    "rnk": {
        "name": "rankinite",
        "group": "Other orthosilicates",
        "formula": "Ca3Si2O7",
        "crystal_system": "",
        "other_info": "",
    },
    "tit": {
        "name": "titanite",
        "group": "Other orthosilicates",
        "formula": "CaTiSiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "sph": {
        "name": "sphene",
        "group": "Other orthosilicates",
        "formula": "CaTiSiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "spu": {
        "name": "spurrite",
        "group": "Other orthosilicates",
        "formula": "Ca5Si2CO11",
        "crystal_system": "",
        "other_info": "",
    },
    "ty": {
        "name": "tilleyite",
        "group": "Other orthosilicates",
        "formula": "Ca5Si2C2O13",
        "crystal_system": "",
        "other_info": "",
    },
    "zrc": {
        "name": "zircon",
        "group": "Other orthosilicates",
        "formula": "ZrSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "cz": {
        "name": "clinozoisite",
        "group": "Sorosilicates",
        "formula": "Ca2Al3Si3O12(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "ep": {
        "name": "epidote [ordered]",
        "group": "Sorosilicates",
        "formula": "Ca2FeAl2Si3O12(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "fep": {
        "name": "Fe-epidote",
        "group": "Sorosilicates",
        "formula": "Ca2Fe2AlSi3O12(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "law": {
        "name": "lawsonite",
        "group": "Sorosilicates",
        "formula": "CaAl2Si2O6(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "pmt": {
        "name": "piemontite [ordered]",
        "group": "Sorosilicates",
        "formula": "Ca2MnAl2Si3O12(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "zo": {
        "name": "zoisite",
        "group": "Sorosilicates",
        "formula": "Ca2Al3Si3O12(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "vsv": {
        "name": "vesuvianite",
        "group": "Sorosilicates",
        "formula": "Ca19Mg2Al11Si18O69(OH)9",
        "crystal_system": "",
        "other_info": "",
    },
    "crd": {
        "name": "cordierite",
        "group": "Cyclosilicates",
        "formula": "Mg2Al4Si5O18",
        "crystal_system": "",
        "other_info": "",
    },
    "hcrd": {
        "name": "hydrous-cordierite",
        "group": "Cyclosilicates",
        "formula": "Mg2Al4Si5O17(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fcrd": {
        "name": "Fe-cordierite",
        "group": "Cyclosilicates",
        "formula": "Fe2Al4Si5O18",
        "crystal_system": "",
        "other_info": "",
    },
    "mncrd": {
        "name": "Mn-cordierite",
        "group": "Cyclosilicates",
        "formula": "Mn2Al4Si5O18",
        "crystal_system": "",
        "other_info": "",
    },
    "osm1": {
        "name": "osumilite [I]",
        "group": "Cyclosilicates",
        "formula": "KMg2Al5Si10O30",
        "crystal_system": "",
        "other_info": "",
    },
    "osm2": {
        "name": "osumilite [II]",
        "group": "Cyclosilicates",
        "formula": "KMg3Al3Si11O30",
        "crystal_system": "",
        "other_info": "",
    },
    "fosm": {
        "name": "Fe-osumilite",
        "group": "Cyclosilicates",
        "formula": "KFe2Al5Si10O30",
        "crystal_system": "",
        "other_info": "",
    },
    "apv": {
        "name": "Al-perovskite",
        "group": "High-P phases",
        "formula": "AlAlO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cpv": {
        "name": "Ca-perovskite",
        "group": "High-P phases",
        "formula": "CaSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cstn": {
        "name": "CaSi-titanite",
        "group": "High-P phases",
        "formula": "CaSi2O5",
        "crystal_system": "",
        "other_info": "",
    },
    "fak": {
        "name": "Fe-akimotoite",
        "group": "High-P phases",
        "formula": "FeSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "fpv": {
        "name": "Fe-perovskite",
        "group": "High-P phases",
        "formula": "FeSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "frw": {
        "name": "Fe-ringwoodite",
        "group": "High-P phases",
        "formula": "Fe2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "fwd": {
        "name": "Fe-wadsleyite",
        "group": "High-P phases",
        "formula": "Fe2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "mak": {
        "name": "akimotoite",
        "group": "High-P phases",
        "formula": "MgSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "mpv": {
        "name": "Mg-perovskite",
        "group": "High-P phases",
        "formula": "MgSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "mrw": {
        "name": "Mg-ringwoodite",
        "group": "High-P phases",
        "formula": "Mg2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "mwd": {
        "name": "Mg-wadsleyite",
        "group": "High-P phases",
        "formula": "Mg2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "phA": {
        "name": "phaseA",
        "group": "High-P phases",
        "formula": "Mg7Si2O8(OH)6",
        "crystal_system": "",
        "other_info": "",
    },
    "acm": {
        "name": "acmite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "NaFeSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "caes": {
        "name": "Ca-eskola pyroxene",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Ca0.5AlSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "cats": {
        "name": "Ca-tschermaks pyroxene",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaAl2SiO6",
        "crystal_system": "",
        "other_info": "",
    },
    "cen": {
        "name": "clino enstatite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Mg2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "di": {
        "name": "diopside",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaMgSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "en": {
        "name": "enstatite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Mg2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "fs": {
        "name": "ferrosilite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Fe2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "cfs": {
        "name": "clino ferrosilite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Fe2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "hed": {
        "name": "hedenbergite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaFeSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "hen": {
        "name": "Hi-P clinoenstatite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Mg2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "jd": {
        "name": "jadeite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "NaAlSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "kos": {
        "name": "kosmochlor",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "NaCrSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "mgts": {
        "name": "Mg-tschermaks pyroxene",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "MgAl2SiO6",
        "crystal_system": "",
        "other_info": "",
    },
    "pren": {
        "name": "protoenstatite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "Mg2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "pswo": {
        "name": "pseudowollastonite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "pxmn": {
        "name": "pyroxmangite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "MnSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "rhod": {
        "name": "rhodonite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "MnSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "wal": {
        "name": "walstromite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "wo": {
        "name": "wollastonite",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CaSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "anth": {
        "name": "anthophyllite",
        "group": "Amphibole",
        "formula": "Mg7Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "cumm": {
        "name": "cummingtonite",
        "group": "Amphibole",
        "formula": "Mg7Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fact": {
        "name": "ferroactinolite",
        "group": "Amphibole",
        "formula": "Ca2Fe5Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fanth": {
        "name": "Fe-anthophyllite",
        "group": "Amphibole",
        "formula": "Fe7Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fgl": {
        "name": "ferroglaucophane",
        "group": "Amphibole",
        "formula": "Na2Fe3Al2Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "gl": {
        "name": "glaucophane",
        "group": "Amphibole",
        "formula": "Na2Mg3Al2Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "grun": {
        "name": "grunerite",
        "group": "Amphibole",
        "formula": "Fe7Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "parg": {
        "name": "pargasite",
        "group": "Amphibole",
        "formula": "NaCa2Mg4Al3Si6O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "rieb": {
        "name": "riebeckite",
        "group": "Amphibole",
        "formula": "Na2Fe5Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "tr": {
        "name": "tremolite",
        "group": "Amphibole",
        "formula": "Ca2Mg5Si8O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "ts": {
        "name": "tschermakite",
        "group": "Amphibole",
        "formula": "Ca2Mg3Al4Si6O22(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "deer": {
        "name": "deerite",
        "group": "Other chain silicates",
        "formula": "Fe18Si12O40(OH)10",
        "crystal_system": "",
        "other_info": "",
    },
    "fcar": {
        "name": "ferrocarpholite",
        "group": "Other chain silicates",
        "formula": "FeAl2Si2O6(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "fspr": {
        "name": "Fe-sapphirine [221]",
        "group": "Other chain silicates",
        "formula": "Fe4Al8Si2O20",
        "crystal_system": "",
        "other_info": "",
    },
    "mcar": {
        "name": "magnesiocarpholite",
        "group": "Other chain silicates",
        "formula": "MgAl2Si2O6(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "spr4": {
        "name": "sapphirine [221]",
        "group": "Other chain silicates",
        "formula": "Mg4Al8Si2O20",
        "crystal_system": "",
        "other_info": "",
    },
    "spr5": {
        "name": "sapphirine [351]",
        "group": "Other chain silicates",
        "formula": "Mg3Al10SiO20",
        "crystal_system": "",
        "other_info": "",
    },
    "ann": {
        "name": "annite",
        "group": "Micas",
        "formula": "KFe3AlSi3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "cel": {
        "name": "celadonite",
        "group": "Micas",
        "formula": "KMgAlSi4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "east": {
        "name": "eastonite",
        "group": "Micas",
        "formula": "KMg2Al3Si2O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fcel": {
        "name": "ferroceladonite",
        "group": "Micas",
        "formula": "KFeAlSi4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "ma": {
        "name": "margarite",
        "group": "Micas",
        "formula": "CaAl4Si2O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mnbi": {
        "name": "Mn-biotite",
        "group": "Micas",
        "formula": "KMn3AlSi3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mu": {
        "name": "muscovite",
        "group": "Micas",
        "formula": "KAl3Si3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "naph": {
        "name": "sodaphlogopite",
        "group": "Micas",
        "formula": "NaMg3AlSi3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "pa": {
        "name": "paragonite",
        "group": "Micas",
        "formula": "NaAl3Si3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "phl": {
        "name": "phlogopite",
        "group": "Micas",
        "formula": "KMg3AlSi3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "afchl": {
        "name": "Al-free chlorite",
        "group": "Chlorites",
        "formula": "Mg6Si4O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "ames": {
        "name": "amesite [14A]",
        "group": "Chlorites",
        "formula": "Mg4Al4Si2O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "clin": {
        "name": "clinochlore [ordered]",
        "group": "Chlorites",
        "formula": "Mg5Al2Si3O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "daph": {
        "name": "daphnite",
        "group": "Chlorites",
        "formula": "Fe5Al2Si3O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "fsud": {
        "name": "ferrosudoite",
        "group": "Chlorites",
        "formula": "Fe2Al4Si3O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "mnchl": {
        "name": "Mn-chlorite",
        "group": "Chlorites",
        "formula": "Mn5Al2Si3O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "sud": {
        "name": "sudoite",
        "group": "Chlorites",
        "formula": "Mg2Al4Si3O10(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "atg": {
        "name": "antigorite",
        "group": "Other sheet silicates",
        "formula": "Mg48Si34O85(OH)62",
        "crystal_system": "",
        "other_info": "",
    },
    "chr": {
        "name": "chrysotile",
        "group": "Other sheet silicates",
        "formula": "Mg3Si2O5(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "fpre": {
        "name": "ferri-prehnite",
        "group": "Other sheet silicates",
        "formula": "Ca2FeAlSi3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "fstp": {
        "name": "ferrostilpnomelane",
        "group": "Other sheet silicates",
        "formula": "K0.5Fe5Al2Si8O18(OH)12.5",
        "crystal_system": "",
        "other_info": "",
    },
    "fta": {
        "name": "ferrotalc",
        "group": "Other sheet silicates",
        "formula": "Fe3Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "glt": {
        "name": "greenalite",
        "group": "Other sheet silicates",
        "formula": "Fe3Si2O5(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "kao": {
        "name": "kaolinite",
        "group": "Other sheet silicates",
        "formula": "Al2Si2O5(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "liz": {
        "name": "lizardite",
        "group": "Other sheet silicates",
        "formula": "Mg3Si2O5(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "minm": {
        "name": "Mg-minnesotaite",
        "group": "Other sheet silicates",
        "formula": "Mg3Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "minn": {
        "name": "minnesotaite",
        "group": "Other sheet silicates",
        "formula": "Fe3Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mstp": {
        "name": "Mg-stilpnomelane",
        "group": "Other sheet silicates",
        "formula": "K0.5Mg5Al2Si8O18(OH)12.5",
        "crystal_system": "",
        "other_info": "",
    },
    "pre": {
        "name": "prehnite",
        "group": "Other sheet silicates",
        "formula": "Ca2Al2Si3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "prl": {
        "name": "pyrophyllite",
        "group": "Other sheet silicates",
        "formula": "Al2Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "ta": {
        "name": "talc",
        "group": "Other sheet silicates",
        "formula": "Mg3Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "tap": {
        "name": "prl-talc",
        "group": "Other sheet silicates",
        "formula": "Al2Si4O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "tats": {
        "name": "tschermak-talc",
        "group": "Other sheet silicates",
        "formula": "Mg2Al2Si3O10(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "abh": {
        "name": "albite [high]",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "ab": {
        "name": "albite",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "an": {
        "name": "anorthite",
        "group": "Feldspars & feldspathoid",
        "formula": "CaAl2Si2O8",
        "crystal_system": "",
        "other_info": "",
    },
    "anl": {
        "name": "analcite",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSi2O5(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "cg": {
        "name": "carnegieite [low]",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "cgh": {
        "name": "carnegieite [high]",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "kcm": {
        "name": "K-cymrite",
        "group": "Feldspars & feldspathoid",
        "formula": "KAlSi3O7(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "kls": {
        "name": "kalsilite",
        "group": "Feldspars & feldspathoid",
        "formula": "KAlSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "lc": {
        "name": "leucite",
        "group": "Feldspars & feldspathoid",
        "formula": "KAlSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "mic": {
        "name": "microcline",
        "group": "Feldspars & feldspathoid",
        "formula": "KAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "ne": {
        "name": "nepheline",
        "group": "Feldspars & feldspathoid",
        "formula": "NaAlSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "san": {
        "name": "sanidine",
        "group": "Feldspars & feldspathoid",
        "formula": "KAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "coe": {
        "name": "coesite",
        "group": "Silica minerals",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "crst": {
        "name": "cristobalite [high]",
        "group": "Silica minerals",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "q": {
        "name": "quartz",
        "group": "Silica minerals",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "stv": {
        "name": "stishovite",
        "group": "Silica minerals",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "trd": {
        "name": "tridymite [high]",
        "group": "Silica minerals",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "heu": {
        "name": "heulandite",
        "group": "Other framework silicates",
        "formula": "CaAl2Si7O12(OH)12",
        "crystal_system": "",
        "other_info": "",
    },
    "hol": {
        "name": "hollandite",
        "group": "Other framework silicates",
        "formula": "KAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "lmt": {
        "name": "laumontite",
        "group": "Other framework silicates",
        "formula": "CaAl2Si4O8(OH)8",
        "crystal_system": "",
        "other_info": "",
    },
    "me": {
        "name": "meionite",
        "group": "Other framework silicates",
        "formula": "Ca4Al6Si6CO27",
        "crystal_system": "",
        "other_info": "",
    },
    "sdl": {
        "name": "sodalite",
        "group": "Other framework silicates",
        "formula": "Na8Al6Si6O24Cl2",
        "crystal_system": "",
        "other_info": "",
    },
    "stlb": {
        "name": "stilbite",
        "group": "Other framework silicates",
        "formula": "CaAl2Si7O11(OH)14",
        "crystal_system": "",
        "other_info": "",
    },
    "wa": {
        "name": "Si-wadeite",
        "group": "Other framework silicates",
        "formula": "K2Si4O9",
        "crystal_system": "",
        "other_info": "",
    },
    "wrk": {
        "name": "wairakite",
        "group": "Other framework silicates",
        "formula": "CaAl2Si4O10(OH)4",
        "crystal_system": "",
        "other_info": "",
    },
    "bdy": {
        "name": "baddeleyite",
        "group": "Oxides",
        "formula": "ZrO2",
        "crystal_system": "",
        "other_info": "",
    },
    "bix": {
        "name": "bixbyite",
        "group": "Oxides",
        "formula": "Mn2O3",
        "crystal_system": "",
        "other_info": "",
    },
    "cor": {
        "name": "corundum",
        "group": "Oxides",
        "formula": "Al2O3",
        "crystal_system": "",
        "other_info": "",
    },
    "cup": {
        "name": "cuprite",
        "group": "Oxides",
        "formula": "Cu2O",
        "crystal_system": "",
        "other_info": "",
    },
    "esk": {
        "name": "eskolaite",
        "group": "Oxides",
        "formula": "Cr2O3",
        "crystal_system": "",
        "other_info": "",
    },
    "fper": {
        "name": "ferropericlase",
        "group": "Oxides",
        "formula": "FeO",
        "crystal_system": "",
        "other_info": "",
    },
    "geik": {
        "name": "geikielite",
        "group": "Oxides",
        "formula": "MgTiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "hem": {
        "name": "hematite",
        "group": "Oxides",
        "formula": "Fe2O3",
        "crystal_system": "",
        "other_info": "",
    },
    "herc": {
        "name": "hercynite",
        "group": "Oxides",
        "formula": "FeAl2O4",
        "crystal_system": "",
        "other_info": "",
    },
    "ilm": {
        "name": "ilmenite",
        "group": "Oxides",
        "formula": "FeTiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "lime": {
        "name": "lime",
        "group": "Oxides",
        "formula": "CaO",
        "crystal_system": "",
        "other_info": "",
    },
    "mang": {
        "name": "manganosite",
        "group": "Oxides",
        "formula": "MnO",
        "crystal_system": "",
        "other_info": "",
    },
    "mcor": {
        "name": "MgSi-corundum",
        "group": "Oxides",
        "formula": "MgSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "mft": {
        "name": "magnesioferrite",
        "group": "Oxides",
        "formula": "MgFe2O4",
        "crystal_system": "",
        "other_info": "",
    },
    "mt": {
        "name": "magnetite",
        "group": "Oxides",
        "formula": "Fe3O4",
        "crystal_system": "",
        "other_info": "",
    },
    "GeO2": {
        "name": "germanium oxide",
        "group": "Oxides",
        "formula": "GeO2",
        "crystal_system": "",
        "other_info": "",
    },
    "NiO": {
        "name": "nickel oxide",
        "group": "Oxides",
        "formula": "NiO",
        "crystal_system": "",
        "other_info": "",
    },
    "per": {
        "name": "periclase",
        "group": "Oxides",
        "formula": "MgO",
        "crystal_system": "",
        "other_info": "",
    },
    "picr": {
        "name": "picrochromite",
        "group": "Oxides",
        "formula": "MgCr2O4",
        "crystal_system": "",
        "other_info": "",
    },
    "pnt": {
        "name": "pyrophanite",
        "group": "Oxides",
        "formula": "MnTiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "pntpv": {
        "name": "pyrophanite [perovskite]",
        "group": "High-P phases",
        "formula": "MnTiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "ru": {
        "name": "rutile",
        "group": "Oxides",
        "formula": "TiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "sp": {
        "name": "spinel",
        "group": "Oxides",
        "formula": "MgAl2O4",
        "crystal_system": "",
        "other_info": "",
    },
    "ten": {
        "name": "tenorite",
        "group": "Oxides",
        "formula": "CuO",
        "crystal_system": "",
        "other_info": "",
    },
    "usp": {
        "name": "ulvospinel",
        "group": "Oxides",
        "formula": "Fe2TiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "br": {
        "name": "brucite",
        "group": "Hydroxides",
        "formula": "Mg(OH)2",
        "crystal_system": "",
        "other_info": "",
    },
    "dsp": {
        "name": "diaspore",
        "group": "Hydroxides",
        "formula": "AlO(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "gth": {
        "name": "goethite",
        "group": "Hydroxides",
        "formula": "FeO(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "ank": {
        "name": "ankerite",
        "group": "Carbonates",
        "formula": "CaFe(CO3)2",
        "crystal_system": "",
        "other_info": "",
    },
    "arag": {
        "name": "aragonite",
        "group": "Carbonates",
        "formula": "CaCO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cc": {
        "name": "calcite",
        "group": "Carbonates",
        "formula": "CaCO3",
        "crystal_system": "",
        "other_info": "",
    },
    "dol": {
        "name": "dolomite",
        "group": "Carbonates",
        "formula": "CaMg(CO3)2",
        "crystal_system": "",
        "other_info": "",
    },
    "mag": {
        "name": "magnesite",
        "group": "Carbonates",
        "formula": "MgCO3",
        "crystal_system": "",
        "other_info": "",
    },
    "rhc": {
        "name": "rhodochrosite",
        "group": "Carbonates",
        "formula": "MnCO3",
        "crystal_system": "",
        "other_info": "",
    },
    "sid": {
        "name": "siderite",
        "group": "Carbonates",
        "formula": "FeCO3",
        "crystal_system": "",
        "other_info": "",
    },
    "any": {
        "name": "anhydrite",
        "group": "Halides & sulphides",
        "formula": "CaSO4",
        "crystal_system": "",
        "other_info": "",
    },
    "hlt": {
        "name": "halite",
        "group": "Halides & sulphides",
        "formula": "NaCl",
        "crystal_system": "",
        "other_info": "",
    },
    "lot": {
        "name": "low troilite",
        "group": "Halides & sulphides",
        "formula": "FeS",
        "crystal_system": "",
        "other_info": "",
    },
    "pyr": {
        "name": "pyrite",
        "group": "Halides & sulphides",
        "formula": "FeS2",
        "crystal_system": "",
        "other_info": "",
    },
    "syv": {
        "name": "sylvite",
        "group": "Halides & sulphides",
        "formula": "KCl",
        "crystal_system": "",
        "other_info": "",
    },
    "tro": {
        "name": "troilite",
        "group": "Halides & sulphides",
        "formula": "FeS",
        "crystal_system": "",
        "other_info": "",
    },
    "trot": {
        "name": "pyrrhotite",
        "group": "Halides & sulphides",
        "formula": "FeS",
        "crystal_system": "",
        "other_info": "",
    },
    "trov": {
        "name": "pyrrhotite",
        "group": "Halides & sulphides",
        "formula": "Fe0.875S",
        "crystal_system": "",
        "other_info": "",
    },
    "Cu": {
        "name": "copper",
        "group": "Elements",
        "formula": "Cu",
        "crystal_system": "",
        "other_info": "",
    },
    "Pt": {
        "name": "platinum",
        "group": "Elements",
        "formula": "Pt",
        "crystal_system": "",
        "other_info": "",
    },
    "Zn": {
        "name": "zinc",
        "group": "Elements",
        "formula": "Zn",
        "crystal_system": "",
        "other_info": "",
    },
    "Pb": {
        "name": "lead",
        "group": "Elements",
        "formula": "Pb",
        "crystal_system": "",
        "other_info": "",
    },
    "Ag": {
        "name": "silver",
        "group": "Elements",
        "formula": "Ag",
        "crystal_system": "",
        "other_info": "",
    },
    "Au": {
        "name": "gold",
        "group": "Elements",
        "formula": "Au",
        "crystal_system": "",
        "other_info": "",
    },
    "Ge": {
        "name": "germanium",
        "group": "Elements",
        "formula": "Ge",
        "crystal_system": "",
        "other_info": "",
    },
    "diam": {
        "name": "diamond",
        "group": "Elements",
        "formula": "C",
        "crystal_system": "",
        "other_info": "",
    },
    "gph": {
        "name": "graphite",
        "group": "Elements",
        "formula": "C",
        "crystal_system": "",
        "other_info": "",
    },
    "iron": {
        "name": "iron",
        "group": "Elements",
        "formula": "Fe",
        "crystal_system": "",
        "other_info": "",
    },
    "Ni": {
        "name": "nickel",
        "group": "Elements",
        "formula": "Ni",
        "crystal_system": "",
        "other_info": "",
    },
    "S": {
        "name": "sulphur",
        "group": "Elements",
        "formula": "S",
        "crystal_system": "",
        "other_info": "",
    },
    "H2O": {
        "name": "water",
        "group": "Gas species",
        "formula": "H2O",
        "crystal_system": "",
        "other_info": "",
    },
    "CO2": {
        "name": "carbon dioxide",
        "group": "Gas species",
        "formula": "CO2",
        "crystal_system": "",
        "other_info": "",
    },
    "CO": {
        "name": "carbon monoxide",
        "group": "Gas species",
        "formula": "CO",
        "crystal_system": "",
        "other_info": "",
    },
    "CH4": {
        "name": "methane",
        "group": "Gas species",
        "formula": "CH4",
        "crystal_system": "",
        "other_info": "",
    },
    "O2": {
        "name": "oxygen",
        "group": "Gas species",
        "formula": "O2",
        "crystal_system": "",
        "other_info": "",
    },
    "H2": {
        "name": "hydrogen",
        "group": "Gas species",
        "formula": "H2",
        "crystal_system": "",
        "other_info": "",
    },
    "S2": {
        "name": "sulphur gas",
        "group": "Gas species",
        "formula": "S2",
        "crystal_system": "",
        "other_info": "",
    },
    "H2S": {
        "name": "hydrogen sulphide",
        "group": "Gas species",
        "formula": "H2S",
        "crystal_system": "",
        "other_info": "",
    },
    "abL": {
        "name": "albite liquid",
        "group": "Melt species",
        "formula": "NaAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "anL": {
        "name": "anorthite liquid",
        "group": "Melt species",
        "formula": "CaAl2Si2O8",
        "crystal_system": "",
        "other_info": "",
    },
    "corL": {
        "name": "corundum liquid",
        "group": "Melt species",
        "formula": "Al2O3",
        "crystal_system": "",
        "other_info": "",
    },
    "diL": {
        "name": "diopside liquid",
        "group": "Melt species",
        "formula": "CaMgSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "enL": {
        "name": "enstatite liquid",
        "group": "Melt species",
        "formula": "Mg2Si2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "faL": {
        "name": "fayalite liquid",
        "group": "Melt species",
        "formula": "Fe2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "foL": {
        "name": "forsterite liquid",
        "group": "Melt species",
        "formula": "Mg2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "h2oL": {
        "name": "H2O liquid",
        "group": "Melt species",
        "formula": "H2O",
        "crystal_system": "",
        "other_info": "",
    },
    "hltL": {
        "name": "halite liquid",
        "group": "Melt species",
        "formula": "NaCl",
        "crystal_system": "",
        "other_info": "",
    },
    "kspL": {
        "name": "K-feldspar liquid",
        "group": "Melt species",
        "formula": "KAlSi3O8",
        "crystal_system": "",
        "other_info": "",
    },
    "lcL": {
        "name": "leucite liquid",
        "group": "Melt species",
        "formula": "KAlSi2O6",
        "crystal_system": "",
        "other_info": "",
    },
    "limL": {
        "name": "CaO liquid",
        "group": "Melt species",
        "formula": "CaO",
        "crystal_system": "",
        "other_info": "",
    },
    "neL": {
        "name": "nepheline liquid",
        "group": "Melt species",
        "formula": "NaAlSiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "perL": {
        "name": "MgO liquid",
        "group": "Melt species",
        "formula": "MgO",
        "crystal_system": "",
        "other_info": "",
    },
    "qL": {
        "name": "quartz liquid",
        "group": "Melt species",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "silL": {
        "name": "sillimanite liquid",
        "group": "Melt species",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "syvL": {
        "name": "sylvite liquid",
        "group": "Melt species",
        "formula": "KCl",
        "crystal_system": "",
        "other_info": "",
    },
    "woL": {
        "name": "wollastonite liquid",
        "group": "Melt species",
        "formula": "CaSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "melt": {
        "name": "silicate melt",
        "group": "Melt species",
        "formula": "melt",
        "crystal_system": "",
        "other_info": "",
    },
    "H+": {
        "name": "hydrogen ion",
        "group": "Aqueous species",
        "formula": "H+",
        "crystal_system": "",
        "other_info": "",
    },
    "Cl-": {
        "name": "chloride ion",
        "group": "Aqueous species",
        "formula": "Cl−",
        "crystal_system": "",
        "other_info": "",
    },
    "OH-": {
        "name": "hydroxyl ion",
        "group": "Aqueous species",
        "formula": "HO−",
        "crystal_system": "",
        "other_info": "",
    },
    "Na+": {
        "name": "sodium ion",
        "group": "Aqueous species",
        "formula": "Na+",
        "crystal_system": "",
        "other_info": "",
    },
    "K+": {
        "name": "potassium ion",
        "group": "Aqueous species",
        "formula": "K+",
        "crystal_system": "",
        "other_info": "",
    },
    "Ca++": {
        "name": "calcium ion",
        "group": "Aqueous species",
        "formula": "Ca++",
        "crystal_system": "",
        "other_info": "",
    },
    "Mg++": {
        "name": "magnesium ion",
        "group": "Aqueous species",
        "formula": "Mg++",
        "crystal_system": "",
        "other_info": "",
    },
    "Fe++": {
        "name": "ferrous ion",
        "group": "Aqueous species",
        "formula": "Fe++",
        "crystal_system": "",
        "other_info": "",
    },
    "Al+++": {
        "name": "aluminium ion",
        "group": "Aqueous species",
        "formula": "Al+++",
        "crystal_system": "",
        "other_info": "",
    },
    "CO3-": {
        "name": "carbonate ion",
        "group": "Aqueous species",
        "formula": "CO3--",
        "crystal_system": "",
        "other_info": "",
    },
    "AlOH3": {
        "name": "aluminium hydroxide",
        "group": "Aqueous species",
        "formula": "Al(OH)3",
        "crystal_system": "",
        "other_info": "",
    },
    "AlOH4-": {
        "name": "aluminium hydroxide",
        "group": "Aqueous species",
        "formula": "Al(OH)4-",
        "crystal_system": "",
        "other_info": "",
    },
    "KOH": {
        "name": "potassium hydroxide",
        "group": "Aqueous species",
        "formula": "K(OH)",
        "crystal_system": "",
        "other_info": "",
    },
    "HCl": {
        "name": "hydrogen chloride",
        "group": "Aqueous species",
        "formula": "HCl",
        "crystal_system": "",
        "other_info": "",
    },
    "KCl": {
        "name": "potassium chloride",
        "group": "Aqueous species",
        "formula": "KCl",
        "crystal_system": "",
        "other_info": "",
    },
    "NaCl": {
        "name": "sodium chloride",
        "group": "Aqueous species",
        "formula": "NaCl",
        "crystal_system": "",
        "other_info": "",
    },
    "CaCl2": {
        "name": "calcium chloride",
        "group": "Aqueous species",
        "formula": "CaCl2",
        "crystal_system": "",
        "other_info": "",
    },
    "CaCl+": {
        "name": "calcium chloride",
        "group": "Aqueous species",
        "formula": "CaCl+",
        "crystal_system": "",
        "other_info": "",
    },
    "MgCl2": {
        "name": "magnesium chloride",
        "group": "Aqueous species",
        "formula": "MgCl2",
        "crystal_system": "",
        "other_info": "",
    },
    "MgCl+": {
        "name": "magnesium chloride",
        "group": "Aqueous species",
        "formula": "MgCl+",
        "crystal_system": "",
        "other_info": "",
    },
    "FeCl2": {
        "name": "ferrous chloride",
        "group": "Aqueous species",
        "formula": "FeCl2",
        "crystal_system": "",
        "other_info": "",
    },
    "aqSi": {
        "name": "silica [aq]",
        "group": "Aqueous species",
        "formula": "SiO2",
        "crystal_system": "",
        "other_info": "",
    },
    "HS-": {
        "name": "sulphide [aq]",
        "group": "Aqueous species",
        "formula": "HS-",
        "crystal_system": "",
        "other_info": "",
    },
    "HSO3-": {
        "name": "sulphite [aq]",
        "group": "Aqueous species",
        "formula": "HSO3-",
        "crystal_system": "",
        "other_info": "",
    },
    "SO4--": {
        "name": "sulphate [aq]",
        "group": "Aqueous species",
        "formula": "SO4--",
        "crystal_system": "",
        "other_info": "",
    },
    "HSO4-": {
        "name": "sulphate2 [aq]",
        "group": "Aqueous species",
        "formula": "HSO4-",
        "crystal_system": "",
        "other_info": "",
    },
    "cdgepx": {
        "name": "cadmium germinate [pyroxene]",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CdGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cdgegt": {
        "name": "cadmium germinate [garnet]",
        "group": "Garnets & olivines",
        "formula": "CdGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cdgeilm": {
        "name": "cadmium germinate [ilmenite]",
        "group": "High-P phases",
        "formula": "CdGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cdgepv": {
        "name": "cadmium germinate [perovskite]",
        "group": "High-P phases",
        "formula": "CdGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cdgewo": {
        "name": "cadmium germinate [wollastonite]",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "CdGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "mggeol": {
        "name": "magnesium germinate [olivine]",
        "group": "Garnets & olivines",
        "formula": "Mg2GeO4",
        "crystal_system": "",
        "other_info": "",
    },
    "mggesp": {
        "name": "magnesium germinate [spinel]",
        "group": "Garnets & olivines",
        "formula": "Mg2GeO4",
        "crystal_system": "",
        "other_info": "",
    },
    "mggecpx": {
        "name": "magnesium germinate [clinopyroxene]",
        "group": "Pyroxenes & pyroxenoids",
        "formula": "MgGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "mggeilm": {
        "name": "magnesium germinate [ilmenite]",
        "group": "High-P phases",
        "formula": "MgGeO3",
        "crystal_system": "",
        "other_info": "",
    },
    "triple": {
        "name": "triple point",
        "group": "Aluminosilicates",
        "formula": "Al2SiO5",
        "crystal_system": "",
        "other_info": "",
    },
    "niol": {
        "name": "nickel silicate [olivine]",
        "group": "High-P phases",
        "formula": "Ni2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "nisp": {
        "name": "nickel silicate [spinel]",
        "group": "High-P phases",
        "formula": "Ni2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "cool": {
        "name": "cobalt silicate [olivine]",
        "group": "High-P phases",
        "formula": "Co2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "cosp": {
        "name": "cobalt silicate [spinel]",
        "group": "High-P phases",
        "formula": "Co2SiO4",
        "crystal_system": "",
        "other_info": "",
    },
    "coopx": {
        "name": "cobalt silicate [orthopyroxene]",
        "group": "High-P phases",
        "formula": "CoSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "cocpx": {
        "name": "cobalt silicate [clinopyroxene]",
        "group": "High-P phases",
        "formula": "CoSiO3",
        "crystal_system": "",
        "other_info": "",
    },
    "PbO2": {
        "name": "lead oxide",
        "group": "Oxides",
        "formula": "PbO2",
        "crystal_system": "",
        "other_info": "",
    },
    "irpv": {
        "name": "calcium-iridium perovskite",
        "group": "High-P phases",
        "formula": "CaIrO3",
        "crystal_system": "",
        "other_info": "",
    },
    "irppv": {
        "name": "calcium-iridium post-perovskite",
        "group": "High-P phases",
        "formula": "CaIrO3",
        "crystal_system": "",
        "other_info": "",
    },
    "nacopv": {
        "name": "sodium-cobalt perovskite",
        "group": "High-P phases",
        "formula": "NaCoF3",
        "crystal_system": "",
        "other_info": "",
    },
    "nacoppv": {
        "name": "sodium-cobalt post-perovskite",
        "group": "High-P phases",
        "formula": "NaCoF3",
        "crystal_system": "",
        "other_info": "",
    },
}



#######################################################
## .0. Load Libraries                            !!! ##
#######################################################
from dataclasses import dataclass
from typing import Any

import numpy as np
import pandas as pd
import plotly.graph_objects as go


#######################################################
## .1. Plotly                                    !!! ##
#######################################################
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@dataclass
class RxnDBPlotter:
    df: pd.DataFrame
    ids: list[str]
    dark_mode: bool = False
    font_size: float = 20

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def __post_init__(self):
        """"""
        if "rxn_color_key" not in self.df.columns:
            raise ValueError(
                "DataFrame must contain 'rxn_color_key' column. Did you use the processor's get_colors_for_filtered_df method?"
            )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def plot(self, temperature_units: str, pressure_units: str) -> go.Figure:
        """
        Plot reaction lines (phase diagram) using plotly.
        """
        required_cols = {
            "unique_id",
            "reaction",
            "reaction_names",
            "reactants",
            "reactant_names",
            "reactant_groups",
            "reactant_formulas",
            "products",
            "product_names",
            "product_groups",
            "product_formulas",
            "type",
            "units_P",
            "units_T",
            "T",
            "T_uncertainty",
            "P",
            "P_uncertainty",
            "plot_type",
            "reference",
        }

        if not required_cols.issubset(self.df.columns):
            missing = required_cols - set(self.df.columns)
            raise ValueError(f"Missing required columns in DataFrame: {missing}")

        if temperature_units == "celcius":
            temperature_units_label = "˚C"
        elif temperature_units == "kelvin":
            temperature_units_label = "K"
        else:
            raise ValueError(f"Unknown temperature unit: {temperature_units}")

        if pressure_units == "gigapascal":
            pressure_units_label = "GPa"
        elif pressure_units == "kilobar":
            pressure_units_label = "kbar"
        else:
            raise ValueError(f"Unknown pressure unit: {pressure_units}")

        fig = go.Figure()

        hovertemplate = (
            "%{customdata[0]}<br>"
            "%{customdata[1]}<br>"
            f"(%{{x:.1f}} {temperature_units_label}, %{{y:.2f}} {pressure_units_label})<br>"
            "%{customdata[2]}<extra></extra>"
        )

        for rid in self.ids:
            d = self.df.query("unique_id == @rid")
            if d.empty:
                continue

            color = d["rxn_color_key"].iloc[0]
            plot_type = d["plot_type"].iloc[0]

            if plot_type == "curve":
                fig.add_trace(
                    go.Scatter(
                        x=d["T"],
                        y=d["P"],
                        mode="lines",
                        line=dict(width=2, color=color),
                        hovertemplate=hovertemplate,
                        customdata=np.stack(
                            (d["reaction"], d["unique_id"], d["type"]), axis=-1
                        ),
                    )
                )
            elif plot_type == "point":
                fig.add_trace(
                    go.Scatter(
                        x=d["T"],
                        y=d["P"],
                        mode="markers",
                        marker=dict(size=8, color=color),
                        error_x=dict(
                            type="data", array=d["T_uncertainty"], visible=True
                        ),
                        error_y=dict(
                            type="data", array=d["P_uncertainty"], visible=True
                        ),
                        hovertemplate=hovertemplate,
                        customdata=np.stack(
                            (d["reaction"], d["unique_id"], d["type"]), axis=-1
                        ),
                    )
                )

        layout_settings = self._configure_layout()

        fig.update_layout(
            xaxis_title=f"Temperature ({temperature_units_label})",
            yaxis_title=f"Pressure ({pressure_units_label})",
            showlegend=False,
            autosize=True,
            **layout_settings,
        )

        return fig

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _configure_layout(self) -> dict[str, Any]:
        """"""
        border_color = "#E5E5E5" if self.dark_mode else "black"
        grid_color = "#999999" if self.dark_mode else "#E5E5E5"
        tick_color = "#E5E5E5" if self.dark_mode else "black"
        label_color = "#E5E5E5" if self.dark_mode else "black"
        plot_bgcolor = "#1D1F21" if self.dark_mode else "#FFF"
        paper_bgcolor = "#1D1F21" if self.dark_mode else "#FFF"
        font_color = "#E5E5E5" if self.dark_mode else "black"
        legend_bgcolor = "#404040" if self.dark_mode else "#FFF"

        return {
            "template": "plotly_dark" if self.dark_mode else "plotly_white",
            "font": {"size": self.font_size, "color": font_color},
            "plot_bgcolor": plot_bgcolor,
            "paper_bgcolor": paper_bgcolor,
            "xaxis": {
                "gridcolor": grid_color,
                "title_font": {"color": label_color},
                "tickfont": {"color": tick_color},
                "showline": True,
                "linecolor": border_color,
                "linewidth": 2,
                "mirror": True,
                "constrain": "range",
            },
            "yaxis": {
                "gridcolor": grid_color,
                "title_font": {"color": label_color},
                "tickfont": {"color": tick_color},
                "showline": True,
                "linecolor": border_color,
                "linewidth": 2,
                "mirror": True,
                "constrain": "range",
            },
            "legend": {
                "font": {"color": font_color},
                "bgcolor": legend_bgcolor,
            },
        }



#######################################################
## .0. Load Libraries                            !!! ##
#######################################################
import pandas as pd
import plotly.graph_objects as go
from faicons import icon_svg
from shiny import App, Inputs, Outputs, Session, reactive, render, ui
from shinywidgets import render_plotly

from rxnDB.data.loader import RxnDBLoader
from rxnDB.data.processor import RxnDBProcessor
from rxnDB.ui import configure_ui
from rxnDB.utils import app_dir
from rxnDB.visualize import RxnDBPlotter

#######################################################
## .1. Init Data                                 !!! ##
#######################################################
try:
    in_data = app_dir / "data" / "cache" / "rxnDB.parquet"
    rxnDB_df = RxnDBLoader.load_parquet(in_data)
    processor = RxnDBProcessor(rxnDB_df)
except FileNotFoundError:
    raise FileNotFoundError(f"Error: Data file not found at {in_data.name}!")
except Exception as e:
    raise RuntimeError(f"Error loading or processing data: {e}!")

#######################################################
## .2. Init UI                                   !!! ##
#######################################################
try:
    app_ui: ui.Tag = configure_ui()
except Exception as e:
    raise RuntimeError(f"Error loading shinyapp UI: {e}!")


#######################################################
## .4. Server Logic                              !!! ##
#######################################################
def server(input: Inputs, output: Outputs, session: Session) -> None:
    """Server logic."""

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Reactive state values
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    _rv_toggle_data_type = reactive.value("all")
    _rv_toggle_similar_reactions = reactive.value("off")

    _rv_selected_chemical_system = reactive.value(["Al", "O", "Si"])
    _rv_selected_phase_abbrevs = reactive.value(set())
    _rv_selected_table_columns = reactive.value(
        ["unique_id", "reaction", "type", "reference"]
    )
    _rv_selected_table_rows = reactive.value([])

    _rv_selected_temperature_units = reactive.value("celcius")
    _rv_selected_pressure_units = reactive.value("gigapascal")

    _rv_group_display_modes = reactive.value({})

    _rv_ui_initialized = reactive.value(False)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Initialization
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _re_initialize_once() -> None:
        """Initialize app state once at startup."""
        if not _rv_ui_initialized():
            _re_initialize_defaults()
            _rv_ui_initialized.set(True)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _re_initialize_defaults() -> None:
        """Initialize default values for groups if not already set."""
        checkbox_groups = processor.get_all_group_names()
        current_display_modes = _rv_group_display_modes().copy()

        changed = False
        for group in checkbox_groups:
            if group not in current_display_modes:
                current_display_modes[group] = "name"
                changed = True

        if changed:
            _rv_group_display_modes.set(current_display_modes)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Helper functions for phase selection management
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _convert_phase_abbrevs_to_selected_boxes(
        phases: set[str], display_mode: str, boxes: list[str]
    ) -> set[str]:
        """Convert phase abbreviations to the current display boxes."""
        if not phases:
            return set()

        selections = set()
        if display_mode == "abbreviation":
            selections = phases.intersection(boxes)
        elif display_mode == "name":
            for abbrev in phases:
                name = set(processor.get_phase_name_from_abbrev(abbrev))
                selections.update(name.intersection(boxes))
        elif display_mode == "formula":
            for abbrev in phases:
                formula = set(processor.get_phase_formula_from_abbrev(abbrev))
                selections.update(formula.intersection(boxes))

        return selections

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def _convert_selected_boxes_to_phase_abbrevs(
        selections: list[str], display_mode: str
    ) -> set[str]:
        """Convert selected boxes back to phase abbreviations."""
        if not selections:
            return set()

        abbrevs = set()
        for box in selections:
            if display_mode == "abbreviation":
                abbrevs.add(box)
            elif display_mode == "name":
                abbrev = processor.get_phase_abbrev_from_name(box)
                if abbrev:
                    abbrevs.update(abbrev)
            elif display_mode == "formula":
                abbrev = processor.get_phase_abbrev_from_formula(box)
                if abbrev:
                    abbrevs.update(abbrev)

        return abbrevs

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Reactive UI components
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @render.ui
    def sidebar_chemical_system_ui() -> ui.Tag:
        """Render sidebar chemical sytem UI."""
        if not _rv_ui_initialized():
            return ui.div("Loading ...")

        components = processor.get_all_chemical_components()

        with reactive.isolate():
            selected_chemical_system = _rv_selected_chemical_system()

        return ui.input_selectize(
            "selected_chemical_system",
            "Filter by chemical system",
            choices={element: element for element in sorted(components)},
            selected=list(selected_chemical_system),
            multiple=True,
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @render.ui
    def sidebar_checkbox_ui() -> ui.Tag:
        """Render sidebar checkbox UI."""
        if not _rv_ui_initialized():
            return ui.div("Loading ...")

        checkbox_groups = processor.get_all_group_names()
        if not checkbox_groups:
            return ui.div("No phase groups available ...")

        with reactive.isolate():
            current_display_modes = _rv_group_display_modes()
            current_phases = _rv_selected_phase_abbrevs()
            current_chemical_system = _rv_selected_chemical_system()

        ui_elements = []
        for group in checkbox_groups:
            # Stable UI IDs
            group_id = processor.get_group_id(group)
            id_radio = f"mode_{group_id}"
            id_boxes = f"boxes_{group_id}"

            display_mode = current_display_modes.get(group, "name")
            boxes = processor.get_grouped_phases(
                group, current_chemical_system, display_mode
            )
            selections = _convert_phase_abbrevs_to_selected_boxes(
                current_phases, display_mode, boxes
            )

            display_mode_ui = ui.input_radio_buttons(
                id_radio,
                "Display Mode",
                choices=["abbreviation", "name", "formula"],
                selected=display_mode,
                inline=False,
            )

            popover_icon = ui.span(
                icon_svg("gear"),
                class_="sidebar-popover-icon",
            )

            popover_ui = ui.popover(
                popover_icon,
                ui.div(
                    display_mode_ui,
                    class_="sidebar-popover-radio-btns",
                ),
                title=f"{group} Settings",
                placement="top",
            )

            checkbox_group_ui = ui.input_checkbox_group(
                id_boxes,
                None,
                choices=sorted(boxes),
                selected=list(selections),
            )

            popover_container = ui.div(popover_ui, class_="sidebar-popover-container")
            panel_container = ui.div(
                checkbox_group_ui, popover_container, class_="sidebar-panel-container"
            )

            ui_elements.append(
                ui.accordion_panel(
                    group,
                    panel_container,
                    value=group,
                )
            )

        return ui.accordion(*ui_elements, id="accordion")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @render.ui
    def table_column_selector_ui():
        """Generate table column selector UI (only re-renders on initialization)."""
        if not _rv_ui_initialized():
            return

        columns = processor.data.columns

        return (
            ui.input_selectize(
                "select_table_columns",
                "Select table columns",
                {col: col for col in columns},
                multiple=True,
            ),
        )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # UI event handlers
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _handle_sidebar_ui_changes():
        """Handle sidebar UI changes."""
        if not _rv_ui_initialized():
            return

        checkbox_groups = processor.get_all_group_names()
        current_display_modes = _rv_group_display_modes().copy()
        current_chemical_system = _rv_selected_chemical_system().copy()

        with reactive.isolate():
            current_selected_phase_abbrevs = _rv_selected_phase_abbrevs()

        new_chemical_system = list(input["selected_chemical_system"]())

        chemical_system_state_change = False
        if new_chemical_system is not None:
            if current_chemical_system != new_chemical_system:
                current_chemical_system = new_chemical_system
                chemical_system_state_change = True

        display_mode_state_change = False
        for group in checkbox_groups:
            group_id = processor.get_group_id(group)
            id_radio = f"mode_{group_id}"

            new_display_mode = input[id_radio]()

            if new_display_mode is not None:
                if current_display_modes.get(group) != new_display_mode:
                    current_display_modes[group] = new_display_mode
                    display_mode_state_change = True

            new_boxes = processor.get_grouped_phases(
                group, new_chemical_system, new_display_mode
            )
            new_selections = _convert_phase_abbrevs_to_selected_boxes(
                current_selected_phase_abbrevs, new_display_mode, new_boxes
            )

            ui.update_checkbox_group(
                id=f"boxes_{group_id}",
                choices=sorted(new_boxes),
                selected=list(new_selections),
            )

        if display_mode_state_change:
            _rv_group_display_modes.set(current_display_modes)
        if chemical_system_state_change:
            _rv_selected_chemical_system.set(current_chemical_system)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _handle_phase_selections():
        """
        Handle phase selection changes from checkbox groups and update the
        central _rv_selected_phase_abbrevs state.
        """
        if not _rv_ui_initialized():
            return

        # Get phase groups
        checkbox_groups = processor.get_all_group_names()

        # Get reactive state values
        with reactive.isolate():
            current_display_modes = _rv_group_display_modes()

        # Check for state change and update central list of selected phases
        newly_selected_phase_abbrevs = set()
        for group in checkbox_groups:
            group_id = processor.get_group_id(group)
            id_boxes = f"boxes_{group_id}"

            selected_boxes = input[id_boxes]()

            if selected_boxes is not None:
                display_mode = current_display_modes.get(group, "name")

                phase_abbrevs = _convert_selected_boxes_to_phase_abbrevs(
                    list(selected_boxes), display_mode
                )
                newly_selected_phase_abbrevs.update(phase_abbrevs)

        # Update the central reactive value if the set of selected abbreviations has changed
        if newly_selected_phase_abbrevs != _rv_selected_phase_abbrevs():
            _rv_selected_phase_abbrevs.set(newly_selected_phase_abbrevs)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Global toggle event listeners
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    @reactive.event(input.toggle_similar_reactions)
    def _re_toggle_similar_reactions() -> None:
        """Toggles similar reactions mode."""
        if not _rv_ui_initialized():
            return

        if _rv_toggle_similar_reactions() == "off":
            _rv_toggle_similar_reactions.set("and")
        elif _rv_toggle_similar_reactions() == "and":
            _rv_toggle_similar_reactions.set("or")
        else:
            _rv_toggle_similar_reactions.set("off")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    @reactive.event(input.toggle_data_type)
    def _re_toggle_data_type() -> None:
        """Toggles data type mode."""
        if not _rv_ui_initialized():
            return

        if _rv_toggle_data_type() == "all":
            _rv_toggle_data_type.set("phase_boundary")
        elif _rv_toggle_data_type() == "phase_boundary":
            _rv_toggle_data_type.set("calibration")
        elif _rv_toggle_data_type() == "calibration":
            _rv_toggle_data_type.set("melting_curve")
        else:
            _rv_toggle_data_type.set("all")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    @reactive.event(input.clear_table_row_selection)
    def _re_clear_table_row_selections() -> None:
        """Clears table selections."""
        if not _rv_ui_initialized():
            return

        _rv_selected_table_rows.set([])

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _re_update_pressure_units():
        """Updates selected pressure units and stores the previous value."""
        if not _rv_ui_initialized():
            return

        selected_pressure_units = input.select_pressure_units()
        if (
            selected_pressure_units
            and selected_pressure_units != _rv_selected_pressure_units()
        ):
            _rv_selected_pressure_units.set(selected_pressure_units)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _re_update_temperature_units():
        """Updates selected temperature units and stores the previous value."""
        if not _rv_ui_initialized():
            return

        selected_temperature_units = input.select_temperature_units()

        if (
            selected_temperature_units
            and selected_temperature_units != _rv_selected_temperature_units()
        ):
            _rv_selected_temperature_units.set(selected_temperature_units)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _update_chemical_system():
        """Handle chemical system changes."""
        if not _rv_ui_initialized():
            return

        selected_chemical_system = list(input.selected_chemical_system())

        if selected_chemical_system:
            _rv_selected_chemical_system.set(selected_chemical_system)
        else:
            _rv_selected_chemical_system.set([])

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Table selection event listeners
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _re_update_table_column_selections() -> None:
        """Updates selected table columns."""
        if not _rv_ui_initialized():
            return

        selected_columns = input.select_table_columns()

        if selected_columns:
            _rv_selected_table_columns.set(list(selected_columns))
        else:
            _rv_selected_table_columns.set(
                ["unique_id", "reaction", "type", "reference"]
            )

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    @reactive.event(input.table_selected_rows)
    def _re_update_table_row_selections() -> None:
        """Updates selected table rows."""
        if not _rv_ui_initialized():
            return

        indices = input.table_selected_rows()

        if indices:
            current_table_df = rc_get_table_data()
            if not current_table_df.empty:
                valid_indices = [i for i in indices if i < len(current_table_df)]
                if valid_indices:
                    ids = current_table_df.iloc[valid_indices]["unique_id"].tolist()
                    _rv_selected_table_rows.set(ids)
                else:
                    _rv_selected_table_rows.set([])
            else:
                _rv_selected_table_rows.set([])
        else:
            _rv_selected_table_rows.set([])

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Reactive calculations for data filtering
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.calc
    def rc_get_table_data() -> pd.DataFrame:
        """Get data for table widget."""
        df = rc_get_filtered_data()
        selected_columns = _rv_selected_table_columns()

        array_columns = [
            "T",
            "P",
            "T_uncertainty",
            "P_uncertainty",
            "lnK",
            "lnk_uncertainty",
        ]

        if not df.empty:
            if any(col in selected_columns for col in array_columns):
                return df[selected_columns].round(2).reset_index(drop=True)
            else:
                return (
                    df[selected_columns]
                    .drop_duplicates(subset="unique_id")
                    .reset_index(drop=True)
                )
        else:
            return pd.DataFrame(columns=selected_columns)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.calc
    def rc_get_plotly_data() -> pd.DataFrame:
        """Get data for Plotly widget."""
        df = rc_get_filtered_data()
        selected_table_rows = _rv_selected_table_rows()
        find_similar_mode = _rv_toggle_similar_reactions()
        data_type = _rv_toggle_data_type()

        if selected_table_rows:
            if find_similar_mode != "off":
                reactants = processor.get_reactant_abbrevs_from_ids(selected_table_rows)
                products = processor.get_product_abbrevs_from_ids(selected_table_rows)

                if reactants or products:
                    df = processor.filter_by_reactants_and_product_abbrevs(
                        list(reactants),
                        list(products),
                        method=str(find_similar_mode),
                    )
                    df = convert_units(df)
                else:
                    df = pd.DataFrame(columns=df.columns)
            else:
                df = df[df["unique_id"].isin(selected_table_rows)]

        if data_type == "all":
            return df
        elif data_type == "phase_boundary":
            return (
                df[df["type"] == "phase_boundary"]
                if "type" in df.columns
                else pd.DataFrame(columns=df.columns)
            )
        elif data_type == "calibration":
            return (
                df[df["type"] == "calibration"]
                if "type" in df.columns
                else pd.DataFrame(columns=df.columns)
            )
        elif data_type == "melting_curve":
            return (
                df[df["type"] == "melting_curve"]
                if "type" in df.columns
                else pd.DataFrame(columns=df.columns)
            )
        else:
            return pd.DataFrame(columns=df.columns)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def convert_units(df: pd.DataFrame) -> pd.DataFrame:
        """"""
        current_temperature_units = _rv_selected_temperature_units()
        current_pressure_units = _rv_selected_pressure_units()

        if current_temperature_units == "celcius":
            df = pd.DataFrame(df.apply(processor.convert_T_to_celcius, axis=1))
        elif current_temperature_units == "kelvin":
            df = pd.DataFrame(df.apply(processor.convert_T_to_kelvin, axis=1))

        if current_pressure_units == "gigapascal":
            df = pd.DataFrame(df.apply(processor.convert_P_to_gigapascal, axis=1))
        elif current_pressure_units == "kilobar":
            df = pd.DataFrame(df.apply(processor.convert_P_to_kbar, axis=1))

        return df

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.calc
    def rc_get_filtered_data() -> pd.DataFrame:
        """Initial filtering based on selected phases and plot type."""
        phases = _rv_selected_phase_abbrevs()

        if phases:
            df = processor.filter_by_reactants_and_product_abbrevs(
                list(phases), list(phases)
            )
        else:
            df = pd.DataFrame(columns=processor.data.columns)
            _ = input.clear_table_row_selection()

        return convert_units(df)

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # Render and update widgets
    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @render.data_frame
    def table() -> render.DataTable:
        """Render table."""
        _ = input.clear_table_row_selection()
        df = rc_get_table_data()

        return render.DataTable(df, height="98%", selection_mode="rows")

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @output
    @render_plotly
    def plotly() -> go.FigureWidget:
        """Render plotly."""
        if not _rv_ui_initialized():
            return go.FigureWidget()

        with reactive.isolate():
            df = rc_get_plotly_data()
            current_temperature_units = _rv_selected_temperature_units()
            current_pressure_units = _rv_selected_pressure_units()
            current_dark_mode = input.dark_mode() == "dark"

        df = processor.add_color_keys(df)
        uids = []
        if not df.empty and "unique_id" in df.columns:
            uids = df["unique_id"].unique().tolist()

        plotter = RxnDBPlotter(df, uids, current_dark_mode)
        fig = plotter.plot(current_temperature_units, current_pressure_units)

        # Create and return the widget
        widget = go.FigureWidget(fig)

        return widget

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @reactive.effect
    def _update_plotly_data() -> None:
        """Update plotly widget data without re-rendering."""
        if not _rv_ui_initialized():
            return

        widget = plotly.widget
        if widget is None:
            return

        df = rc_get_plotly_data()
        current_temperature_units = _rv_selected_temperature_units()
        current_pressure_units = _rv_selected_pressure_units()
        current_dark_mode = input.dark_mode() == "dark"

        df = processor.add_color_keys(df.copy())
        uids = []
        if not df.empty and "unique_id" in df.columns:
            uids = df["unique_id"].unique().tolist()

        plotter = RxnDBPlotter(df, uids, current_dark_mode)
        fig = plotter.plot(current_temperature_units, current_pressure_units)

        # Update widget in place using batch_update to prevent flickering
        with widget.batch_update():
            widget.data = ()
            widget.add_traces(fig.data)
            widget.layout.update(fig.layout)  # type: ignore

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    @render.text
    def plot_settings() -> str:
        """Show plot settings info"""
        data_type = f"Type: {_rv_toggle_data_type()}\n"
        similar_reactions = f"Similar rxns: {_rv_toggle_similar_reactions()}\n"

        selected_rows = _rv_selected_table_rows()
        if selected_rows:
            selections_str = "\n".join(selected_rows)
        else:
            selections_str = "None"

        table_selections = f"Table selections:\n{selections_str}"

        info = data_type + similar_reactions + table_selections

        return info


#######################################################
## .5. Shiny App                                 !!! ##
#######################################################
app: App = App(app_ui, server)



#######################################################
## .0. Load Libraries                            !!! ##
#######################################################
from shiny import ui
from shinywidgets import output_widget

from rxnDB.utils import app_dir


#######################################################
## .1. Shiny App UI                              !!! ##
#######################################################
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def configure_ui() -> ui.Tag:
    """
    Configures the Shiny app user interface
    """
    return ui.page_sidebar(
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Sidebar !!
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        ui.sidebar(
            ui.output_ui("sidebar_chemical_system_ui"),
            ui.output_ui("sidebar_checkbox_ui"),
            title="Phases",
            width=300,
            open={"desktop": "open", "mobile": "closed"},
        ),
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Custom CSS !!
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        ui.head_content(ui.include_css(app_dir / "styles.css")),
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        # Plotly and Table !!
        # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
        ui.navset_card_pill(
            ui.nav_panel(
                "Visualization",
                ui.layout_sidebar(
                    ui.sidebar(
                        ui.tooltip(
                            ui.input_action_button(
                                "toggle_data_type",
                                "Toggle Data Type",
                                class_="popover-btn",
                            ),
                            "Show phase boundaries, calibrations, melting curves, or all",
                        ),
                        ui.tooltip(
                            ui.input_action_button(
                                "toggle_similar_reactions",
                                "Toggle Similar Reactions",
                                class_="popover-btn",
                            ),
                            "Find similar rxn sets by intersection or union (for table selections)",
                        ),
                        ui.input_selectize(
                            "select_temperature_units",
                            "Select temperature units",
                            {"celcius": "celcius", "kelvin": "kelvin"},
                        ),
                        ui.input_selectize(
                            "select_pressure_units",
                            "Select pressure units",
                            {"gigapascal": "gigapascal", "kilobar": "kilobar"},
                        ),
                        ui.output_text_verbatim("plot_settings", placeholder=False),
                        title="Plot Controls",
                        open={"desktop": "open", "mobile": "closed"},
                    ),
                    output_widget("plotly"),
                    fillable=True,
                ),
            ),
            ui.nav_panel(
                "Table",
                ui.page_sidebar(
                    ui.sidebar(
                        ui.input_action_button(
                            "clear_table_row_selection",
                            "Clear Table Selections",
                            class_="popover-btn",
                        ),
                        ui.output_ui("table_column_selector_ui"),
                        title="Table Controls",
                        open={"desktop": "open", "mobile": "closed"},
                    ),
                    ui.output_data_frame("table"),
                    fillable=True,
                ),
            ),
            ui.nav_spacer(),
            ui.nav_menu(
                "Settings",
                ui.nav_control(
                    ui.tooltip(
                        ui.input_dark_mode(id="dark_mode", class_="dark-mode-switch"),
                        "Dark/Light Mode",
                    ),
                ),
            ),
        ),
        title="rxnDB",
        fillable=True,
    )



#######################################################
## .0.              Load Libraries               !!! ##
#######################################################
from unittest.mock import MagicMock

import pytest

from rxnDB.app import app
from rxnDB.ui import configure_ui


#######################################################
## .1.                Fixtures                   !!! ##
#######################################################
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@pytest.fixture
def app_ui():
    """
    Fixture to provide a configured UI object for testing
    """
    return configure_ui()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@pytest.fixture
def mock_input():
    """
    Fixture to create a mock input object for testing the server
    """
    mock = MagicMock()
    mock.reactants.return_value = [
        "aluminosilicate",
        "olivine",
        "spinel",
        "wadsleyite",
        "ringwoodite",
    ]
    mock.products.return_value = [
        "aluminosilicate",
        "olivine",
        "spinel",
        "wadsleyite",
        "ringwoodite",
    ]
    mock.mode.return_value = "light"
    mock.datatable_selected_rows.return_value = []

    return mock


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@pytest.fixture
def mock_output():
    """
    Fixture to create a mock output object for testing the server
    """
    return MagicMock()


# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
@pytest.fixture
def mock_session():
    """
    Fixture to create a mock session object for testing the server
    """
    return MagicMock()


#######################################################
## .2.               Test Suite                  !!! ##
#######################################################
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
class TestApp:
    """
    Test suite for the main components of the Shiny app
    """

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def test_ui_configuration(self, app_ui):
        """
        Just check that UI configuration runs without errors
        """
        assert app_ui is not None

    # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    def test_app_creation(self):
        """
        Test that the app object is properly created
        """
        assert app is not None
        assert hasattr(app, "ui")
        assert hasattr(app, "server")


#######################################################
## .3.          Server Smoke Test                !!! ##
#######################################################
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def test_server_smoke(mock_input, mock_output, mock_session):
    """
    Just test that server runs without errors
    """
    try:
        app.server(mock_input, mock_output, mock_session)
        assert True
    except Exception as e:
        pytest.fail(f"Server initialization failed with error: {e}")
